-- Example/Test suite for CoCoAServer computations

Alias TSL := $ts/ts_lib;
TSL.Initialize();

Define CoCoAServerRegister(Var T, N)
  Str := "TRUE
-------------------------------
";
  T.ExpectedOutput := Str;
  For I := 2 To N Do T.ExpectedOutput := T.ExpectedOutput + Str; EndFor;
  $ts/ts_lib.RegisterTest(T);
EndDefine; -- CoCoAServerRegister
  
-------------------------------
-- TEST
Test := Record[Id = "cocoa5-01",Descr = "ReducedGBasis5x Module posto "];

Test.Input :="
Use Q[x,y,z],PosTo;
M := Module([x,z], [z,y]);
RGB := ReducedGBasis(M);
X := ReducedGBasis5x(M, Record(IsPosTo=True));// PosTo is default
X=       [Vector(z, y),       Vector(x, z),       Vector(0, xy - z^2)];
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-02",Descr = "ReducedGBasis5x Module ToPos "];

Test.Input :="
Use Q[x,y,z],PosTo;
M := Module([x,z], [z,y]);
X := ReducedGBasis5x(M, Record(IsPosTo=False));
X=       [       Vector(z, y),       Vector(x, z)];
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-03",Descr = "GBasis5x param "];

Test.Input :="
M := CoCoA5.AddParamOrdMat(DegRevLexMat(2), 1);
Use Z/(32003)[a, x,y], Ord(M);

I := Ideal((a-1)x+(a^2+a)y, (a+1)x + y);
X := GBasis5x(I, Record(NumParams=1));
X=[x, y];
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-04",Descr = "GBasis5x Module shifts "];

Test.Input :="
Use Q[x,y,z], Weights([1,2,4]);
M := Module([x^2-y,1],[x^4-z,y^2]);
Info5 := Record();
Info5.ModuleShifts := Mat([[0,2]]);
X := GBasis5x(M, Info5);
X = [ Vector(x^2 - y, 1), Vector(y^2 - z, y^2 - x^2 - y)];
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-05",Descr = "ReducedGBasis5 Module OrdMat "];

Test.Input :="
OrdMat := Mat([[1,1,1],[2,1,1],[1,1,0]]); // Ring Grading (first 2 Rows) 
                                          // Plus order (last row)
Use Z/(101)[x,y,z], Ord(OrdMat), ToPos;
M := Module([y-x,0,0], [x,0,z], [0,y^2-z^2,0]);
X := ReducedGBasis5(Module(Gens(M)));
X= [Vector(x,0,z),Vector(y,0,z),Vector(0, 0, xz-yz),Vector(0, y^2-z^2, 0)];
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-06",Descr = "ReducedGBasis5x Module shifts gradingdim2"];

Test.Input :="
OrdMat := Mat([[1,1,1],[2,1,1],[1,1,0]]); // Ring Grading (first 2 Rows) 
                                          // Plus order (last row)
Use Z/(101)[x,y,z], Ord(OrdMat), ToPos;
M := Module([y-x,0,0], [x,0,z], [0,y^2-z^2,0]);
Info5 := Record();
Info5.OrdMat := OrdMat;
Info5.GradingDim := 2;
Info5.ModuleShifts := Mat([[3,1,2],[2,2,5]]); // GrDim rows!!
X := ReducedGBasis5x(M, Info5);
X=[Vector(0, y^2-z^2, 0),Vector(x, 0, z),Vector(y, 0, z),Vector(0, 0, xz - yz)];
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-07",Descr = "Elim5"];

Test.Input :="
Use Q[x,y,z,w[3..5]], Weights([7, 4, 3, 1, 1, 1]);
// Use Z/(7)[x,y,z,w[3..5]], Weights([7, 4, 3, 1, 1, 1]); // exbug

I := Ideal(
	   x - 7413431 w[4]^7 - 9162341 w[3]*w[4]*w[5]^5,
	   y - 6521443 w[4]^4 - 2312257 w[3]^2*w[4]*w[5],
	   z - 5329421 w[4]^3 - 2122414 w[3]*w[5]^2
	   );

E  := Elim([w[3],w[4]], I);
E5 := Elim5([w[3],w[4]], I);
E=E5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-08",Descr = "Intersection5  "];

Test.Input :="
Use Q[x,y,z], Weights(1,2,1);
I := Ideal(xy, z^2);
J := Ideal(yz, x-z);
I5 := Intersection5(I, J);
I4 := Intersection(I, J);
I4 = I5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-09",Descr = "Intersection5 ModuleShifts "];

Test.Input :="
Use Q[x,y,z], Weights(1,2,1);
M := Module([xxy, 0], [xz^2, 0], [-xz + xx, -z + x]);
N := Module([xz^2, 0], [xyz, yz], [-xz + xx, -z + x]);
I5 := Intersection5x(M, N, Record(ModuleShifts = [[1,2]]));
I4 := Intersection(M, N);
I4 = I5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-10",Descr = "Intersection5 Module "];

Test.Input :="
Use Z/(101)[x,y,z,t],PosTo;
M:=Module([(x+y+t)^3,y^3],[(x-y-z)^3,z^3],[x^2-y^2,x^2-z^2],[xyz+yzt+ztx,x^3]);
N:=Module([x^2+xy,y^2]);

I4 := Intersection(M,N);// Long
I5 := Intersection5(M,N);// Long translation
I4 = I5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-11",Descr = "Intersection5x param  "];

Test.Input :="
M := CoCoA5.AddParamOrdMat(DegRevLexMat(2), 1);
Use Z/(32003)[a, x,y], Ord(M);

II := Ideal(x-y);
I := (a-1) * x * II;
J := (a+1) * y * II;
X := Intersection5x(I,J, Record(NumParams=1));
X=x*y*II;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-12",Descr = "Colon5x Module/Module 1 "];

Test.Input :="
Use Z/(101)[x,y,z];
MZ := Module(Vector(x,y));
M := Module([x,y], [x,z]);
Info5 := Record();
Info5.OrdMat := Mat([[1,2,2], [1,0,0], [0,0,-1]]);
Info5.GradingDim := 2;
Info5.ModuleShifts := Mat([[1,0], [0,1]]);
T := Colon5x(M, MZ, Info5);
T = Ideal(1);
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-13",Descr = "Colon5x Module/Module 2 "];

Test.Input :="
Use Z/(101)[x,y,z];
MZ := Module(Vector(x,y));
M := Module([x^2,y^2], [x^2,z^2]);
T := Colon5(M, MZ);
T = Ideal(xy^2 - xz^2);
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-14",Descr = "Saturation5x "];

Test.Input :="
Use Q[x,y,z,t];
M := Ideal(x^5, x^2y^2);
Z := x^5;
T4 := Saturation(M, Ideal(Z));
T5 := Saturation5x(M, Ideal(Z), Record());
-- Ideal(x^2y^2, x^5, 1)
T4 = T5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-15",Descr = "Saturation5x Module/Ideal "];

Test.Input :="
Use Z/(101)[x,y,z];
M := Module([x,y],[x,z]);
Z := x;

Info5 := Record();
Info5.OrdMat := Mat([[1,2,2], [1,0,0], [0,0,-1]]);
Info5.GradingDim := 2;
Info5.ModuleShifts := Mat([[1,0], [0,1]]);
T5 := Saturation5x(M, Ideal(Z), Info5);
--    Module([x, z], [0, y - z], [y - z, 0])
T4 := Saturation(M, Ideal(Z));
T4 = T5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-16",Descr = "Saturation5 Module/Ideal "];

Test.Input :="
M:=Module([x^2,y^2],[x^2,z^2]);
T5 := Saturation5(M, Ideal(Z));
--    Module([x^2, z^2], [0, y^2 - z^2], [xy^2 - xz^2, 0], [y^2- z^2, 0])
T4 := Saturation(M, Ideal(Z));
T4 = T5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-17",Descr = "Saturation5 Module/Ideal "];

Test.Input :="
Use Z/(101)[x,y,z,t,u,v];
M:=Module([xyz+yzt+ztu,tuv+uvx+vxy],[xyzt+yztu+ztux,z^4],[tuv+yzt+uvx,t^2z]);
Z:=y^2;
X := Saturation5(M, Ideal(Z));
T4 := Saturation(M, Ideal(Z));
X = T4;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-18",Descr = "SyzOfGens5x Module "];

Test.Input :="
Use Z/(101)[x,y,z], Weights(1,2,4);
L:=[Vector(x^2,x^2), Vector(xy,xy)];
S5 := SyzOfGens5x(Module(L), Record[]);
--    Module([y, -x])
S4 := SyzOfGens(Module(L));
S4 = S5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-19",Descr = "Syz5  "];

Test.Input :="
Use Z/(101)[x,y,z,t,h];
L := [x+y+z+t,xy+yz+zt+tx,xyz+yzt+ztx+txy,xyzt-h^4];
S5 := SyzOfGens5x(Ideal(L), Record[]);
S4 := Syz(L);
S4 = S5;
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-20",Descr = "GBasis5x float "];

Test.Input :="
Use Q[x[0..3]];
L := [Randomized(DensePoly(4)) | I In 1..2];
--GBasis5x(Ideal(L), Record(FloatPrecision=128));
X := LT5x(Ideal(L), Record(FloatPrecision=128));
X=       Ideal(  x[0]^4,  x[0]^3x[1],  x[0]^2x[1]^3,  x[0]x[1]^5,  x[1]^7);

G0 := GBasis5(Ideal(L));
G128 := GBasis5x(Ideal(L), Record(FloatPrecision=128));
Len(G0) = Len(G128);

LT0 := LT5(Ideal(L));
LT128 := LT5x(Ideal(L), Record(FloatPrecision=128));
LT0 = LT128;
";    CoCoAServerRegister(Test, 3);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-22",Descr = "ReducedGBasis5(I) "];

Test.Input :="
Use Q[t,x,y,z];
I:=Ideal(t^3-x^3,t^5-y^5,t^7-z^7);
X := ReducedGBasis5(I);

X = [t^3 - x^3, t^2x^3 - y^5, x^6 - ty^5, t^2y^5 - z^7, x^3y^5 - tz^7, y^10 - x^3z^7];
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-23",Descr = "Elim5 Module"];

Test.Input :="
Use Q[t,x,y,z];
I:=Ideal(t^3-x^3,t^5-y^5,t^7-z^7);
X := Elim5([t],I);
X = Ideal(-y^10 + x^3z^7, -x^12 + y^5z^7, x^9y^5 - z^14);

M := Module([0,x+y],[x,y+z]);
Info5:=Record[];
X := Elim5([z],M);
X = Module([0, x + y], [x^2 + xy, 0]);
";    CoCoAServerRegister(Test, 2);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-24",Descr = "Colon5 "];

Test.Input :="
Use Q[t,x,y,z];
I:=Ideal(x,y);
J:=Ideal(y,z);
K:=Intersection5(I,J);
X := Colon5(Ideal(x^5y,xy^2,z^2),Ideal(x,y,z));
X = Ideal(z^2, xy^2, x^4yz, x^5y);
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-25",Descr = "Saturation5 "];

Test.Input :="
Use Q[t,x,y,z];
X := Saturation5(Ideal(x^5y,xy^2,z^2),Ideal(x,y,z));
X = Ideal(z^2, xy);
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-26",Descr = "Homogenized5x "];

Test.Input :="
M1:=Mat([[2, 1, 1,0],
         [1, 2, 0,1],
         [0, 1, 0,1],
         [1,0,0,0]]);
Use Q[x,y,a,b],Ord(M1),PosTo;
Info5 := Record();
Info5.GradingDim := 2;
Info5.OrdMat :=M1;
I := Ideal(x+1,y^2+x);
X := Homogenized5x([a,b],I,Info5);
X = Ideal(a^2b + x, xb^3 + y^2, x^2b^2 - y^2a^2, y^2a^4 + x^3b);
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-27",Descr = "Saturation5"];

Test.Input :="
M1:=Mat([[2, 1, 1,0],
         [1, 2, 0,1],
         [0, 1, 0,1],
         [1,0,0,0]]);
Use Q[x,y,a,b],Ord(M1),PosTo;
J:=Ideal(x+a^2b,xb^3+y^2);
X := Saturation5(J,Ideal(ab));
X = Ideal(a^2b + x, xb^3 + y^2, x^2b^2 - y^2a^2, y^2a^4 + x^3b);
";    CoCoAServerRegister(Test, 1);

-------------------------------
-- TEST
Test := Record[Id = "cocoa5-28",Descr = "Homogenized5 "];

Test.Input :="
Use Q[x,y,z,h];
I := Ideal(x^2y+xy+1,xy^2+y^2+1);
X := Homogenized5([h],I);
X = Ideal(x - y, y^3 + y^2h + h^3);
";    CoCoAServerRegister(Test, 1);


-------------------------------
-- TEST
Test := Record[Id = "cocoa5-29",Descr = "GBasis5x multi-param "];

Test.Input :="
M := CoCoA5.AddParamOrdMat(DegRevLexMat(2), 3);
Use Z/(32003)[a,b,c, x,y], Ord(M);

I := Ideal((a-1+b^5)x+(a^2+c)y, (a+b+c+1)x + y);
X := GBasis5x(I, Record(NumParams=3));
X=[x, y];
";    CoCoAServerRegister(Test, 1);


-------------------------------
-- TEST
Test := Record[Id = "cocoa5-30",Descr = "Intersection5x multi-param  "];

Test.Input :="
M := CoCoA5.AddParamOrdMat(DegRevLexMat(2), 2);
Use Z/(32003)[a,b, x,y], Ord(M);

II := Ideal(x-y);
I := (a+b-1) * x * II;
J := (a-b^2+1) * y * II;
X := Intersection5x(I,J, Record(NumParams=2));
X=x*y*II;
";    CoCoAServerRegister(Test, 1);


